{% extends "!layout.html" %}

{% block sidebartitle %}
  {%- set _logo_url = logo_url|default(pathto('_static/' + (logo or ""), 1)) %}
  {%- set _root_doc = root_doc|default(master_doc) %}
  <a href="{{ pathto(_root_doc) }}"{% if not theme_logo_only %} class="icon icon-home"{% endif %}>
    {% if not theme_logo_only %}{{ project }}{% endif %}
    {%- if logo or logo_url %}
      <img src="{{ _logo_url }}" class="logo" alt="{{ _('Logo') }}"/>
    {%- endif %}
  </a>

  {%- if version %}
    <div class="version">
      {{ version }}
    </div>
  {%- endif %}

  {%- include "searchbox.html" %}
{% endblock %}
