/* Restore version display removed in sphinx-rtd-theme 3.x. */
.wy-side-nav-search > div.version {
    margin-top: -.4045em;
    margin-bottom: .809em;
    font-weight: normal;
    color: hsla(0, 0%, 100%, .3);
}

/* Fixes https://github.com/readthedocs/sphinx_rtd_theme/issues/1301#issuecomment-1876120817 */
.py.property {
    display: block !important;
}

/* Better display of multi-line signatures. */

dt.sig > dl > dd {
    margin-bottom: 0px;
}

dt.sig > dl {
    margin-bottom: 0px;
}
