// Copyright (c), ETH Zurich and UNC Chapel Hill.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//
//     * Neither the name of ETH Zurich and UNC Chapel Hill nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#include "colmap/geometry/bbox.h"

#include "colmap/util/logging.h"

namespace colmap {

std::vector<Eigen::AlignedBox3d> ComputeEqualPartsBboxes(
    const Eigen::AlignedBox3d& bbox, const Eigen::Vector3i& split) {
  THROW_CHECK_GT(split(0), 0);
  THROW_CHECK_GT(split(1), 0);
  THROW_CHECK_GT(split(2), 0);

  const Eigen::Vector3d extent = bbox.diagonal();
  const Eigen::Vector3d size(
      extent(0) / split(0), extent(1) / split(1), extent(2) / split(2));

  std::vector<Eigen::AlignedBox3d> bboxes;
  bboxes.reserve(split(0) * split(1) * split(2));
  for (int k = 0; k < split(2); ++k) {
    for (int j = 0; j < split(1); ++j) {
      for (int i = 0; i < split(0); ++i) {
        Eigen::Vector3d min(bbox.min().x() + i * size(0),
                            bbox.min().y() + j * size(1),
                            bbox.min().z() + k * size(2));
        bboxes.emplace_back(min, min + size);
      }
    }
  }

  return bboxes;
}

}  // namespace colmap
